from pydelorean.tree.node import *
from pydelorean import treeify

def get_progressive_expansion(node, **kwargs) -> list:
    
    # BASE CASE: TextNode
    if isinstance(node, TextNode):
        return [(node.corpus, node.corpus)]
    
    if kwargs.get("append", False):
        returned_list = [(node.header, node.corpus)]
    else:
        returned_list = []
    for child in node.children:
        if isinstance(child, HeaderNode):
            child_list = get_progressive_expansion(child, append=True)
            returned_list += [(child.header, para) for header, para in child_list]
        elif isinstance(child, TextNode):
            returned_list.append((node.header, child.corpus))
        
    return returned_list


# text = """
# ---
# author: Keane Moraes
# id: 1
# tags:
# - dune
# - ecology
# - fremen
# - religion
# - desert
# ---

# # Appendices

# The appendix begins with an epigraph by Pardot Kynes in which he considers the kind of existence available when humans increase in number in a finite environment.

# ## Appendix I: The Ecology of Dune.
# This appendix details “the ecology of Dune.” It is heavily focused on the story of Pardot Kynes, Arrakis's first planetologist.

# ## Appendix II: The Religion of Dune.
# Before the coming of [[Muad'Dib]], the Fremen of Arrakis practiced a religion whose roots in the Maometh Saari are there for any scholar to see.

# ## Appendix III: Report on Bene Gesserit Motives and Purposes.
# This appendix details a “report on Bene Gesserit motives and purpose.”
# The narrator introduces it by noting that Lady Jessica commissioned the report directly after the “Arrakis Affair.” 

# ### Narratives
# The document is noted as being extremely honest in tone.

# ## Appendix IV: The Almanak eb-Ashraf (Selected Excerpts of the Noble Houses)
# This appendix details selected excerpts about the noble Houses of Dune. The first entry discusses the Padishah Emperor Shaddam IV of House Corrino. His rule is most significant for the “Arrakis Revolt,” which historians ascribe to his poor court politics. 

# # Terminology of the Imperium

# Hello there

# ## Another One

# asdfas
# """

# forest = treeify("dune_appendices.md", text)

# root = forest.root
# header1 = root.children[0]

# ans = get_progressive_expansion(root, append=True)

# for i in ans:
#     print(i)
#     print("==================================")




class File:
    pass

# FIGUREOUT: Do we really need a file class for each of the supported formats ?
# FIGUREOUT: What more attributes / methods should these classes have ?

class MarkdownFile(File):
    pass

class RestructuredFile(File):
    pass

class AsciidocFile(File):
    pass

class TextFile(File):
    pass

class YAMLFile(File):
    pass

class JSONFile(File):
    pass


FILE_EXTENSIONS = {
    ".md":      MarkdownFile,
    ".rst":     RestructuredFile,
    ".adoc":    AsciidocFile,
    ".txt":     TextFile,
    ".yaml":    YAMLFile,
    ".yml":     YAMLFile,
    ".json":    JSONFile
}


from bs4 import BeautifulSoup
from .delorean import *
from .tree import Forest
import re


# TODO: Check the type of the input and call the appropriate function
# Use the newly defined File classes for this. 
def treeify(name:str, text:str, *args, **kwargs) -> Forest:
    
    return mdtreeify(name, text, *args, **kwargs)


def textify(forest:Forest, *args, **kwargs) -> str:
    
    return mdtextify(forest)


# TODO: I want to only expose treeify and textify to the user. Move this someplace else.
def clean_markdown(md:str) -> str:
        
    # Remove the metadata
    _, markdown_text = find_metadata(md)
    
     # Create a BeautifulSoup object with the input markdown text
    soup = BeautifulSoup(markdown_text, 'html.parser')

    # Remove HTML tags from the document
    markdown_text = soup.get_text()
    
    # Remove links, images, and backlinks
    markdown_text = re.sub(r'!\[(.*?)\]\((.*?)\)', '', markdown_text)
    markdown_text = re.sub(r'\[(.*?)\]\((.*?)\)', r'\1', markdown_text)
    markdown_text = re.sub(r'\[\[(.*?)\]\]', r'\1', markdown_text)
    
    # Remove quotation marks
    markdown_text = markdown_text.replace('>', '')  
    
    # Remove bold and italics formatting
    markdown_text = re.sub(r'\*\*(.*?)\*\*', r'\1', markdown_text)
    markdown_text = re.sub(r'__(.*?)__', r'\1', markdown_text)
    markdown_text = re.sub(r'\*(.*?)\*', r'\1', markdown_text)
    markdown_text = re.sub(r'_(.*?)_', r'\1', markdown_text)
    
    # Remove empty lines
    markdown_text = re.sub(r'^\s*$', '', markdown_text, flags=re.MULTILINE)
    
    return markdown_text
    



"""_summary_

"""

from bigtree import Node

# TODO: Add all the documentation and convert to snake case soon.


class TextNode(Node):
    
    FRACTION_PRINTABLE = 0.25
    
    def __init__(self, name:str, text:str, **kwargs):
        self.text = text
        super().__init__(name, **kwargs)
        
        
    def __getitem__(self, index:int):
        if index < 0 or index >= len(self.text):
            raise IndexError("Index out of bounds")
        return self.text[index]
           
    def __str__(self):
        return self.text[0:min(10, int(self.FRACTION_PRINTABLE * len(self.text)))] + " ..."

    def __repr__(self) -> str:
        return "(Text) " + self.text[0:min(10, int(self.FRACTION_PRINTABLE * len(self.text)))] + " ..."
    
    @property
    def text(self):
        return self._text
    
    @property
    def corpus(self):
        return self._text
    
    @text.setter
    def text(self, text:str):
        self._text = text


class HeaderNode(Node):
    
    def __init__(self, header:str, headerNumber:int, **kwargs):
        self.header = header
        self._header_level = headerNumber
        self._corpus = []
        super().__init__(header, **kwargs)
        
    def __str__(self):
        return self.header
    
    def __repr__(self):
        return f"(h{self._header_level}) {self.header}"
    
    @property
    def header_level(self) -> int:
        return self._header_level
    
    @header_level.setter
    def header_level(self, headerNumber:int) -> None:
        self._header_level = headerNumber

    def __pow__(self, other):
        self._corpus.append(other)
        
    @property
    def corpus(self) -> str:
        final_text = self.header + "\n"
        for child in self._corpus:
            final_text += child.corpus + "\n"
        return final_text    

        



""" This file contains the main functions for the delorean package.
"""
    
import re

import frontmatter
from .files import *
from .tree import *
from .parser import *

# ==================================================================================================
#                                           TREEIFY
# ==================================================================================================

# TODO: Rename to snake_case in the future.


def find_backlinks(input_text:str) -> list:
    """ 
    Function to find all backlinks in a given text.
    """
    
    backlinks = []

    # Regular expression pattern to match backlinks
    pattern = r'\[\[(.*?)\]\]'
    backlinks = re.findall(pattern, input_text)
    return backlinks


def find_tags(input_text:str) -> list:
    """
    Function to find all tags in a given text.
    """
    
    tags = []

    # Regular expression pattern to match backlinks
    pattern = r'#([a-zA-Z0-9_]+)'
    tags = re.findall(pattern, input_text)
    return tags


def find_metadata(input_text:str):
    post = frontmatter.loads(input_text)
    return post.metadata, post.content


def mdtreeify(name:str, md:str, *args, **kwargs) -> MarkdownForest:
    """
    Converts markdown file to a MarkdownForest
    """
    
    meta, cont = find_metadata(md)
    
    backlinks = find_backlinks(cont)
    tags = find_tags(cont)
    
    if cont == "":
        # If the file is empty, just create a root node
        tree = Node("root")
        return tree
    
    
    parser = MarkdownParser(name, cont)
    tree = parser.parse()
    
    returnForest = MarkdownForest(tree, name, metadata=meta)
    
    for backlink in backlinks:
        returnForest.add_backlink(backlink)
    for tag in tags:
        returnForest.add_tag(tag)

    return returnForest


def rsttreeify(name:str, rst:str, *args, **kwargs) -> RestructuredForest:
    pass


def asciidoc_treeify(name:str, asciidoc:str, *args, **kwargs) -> AsciidocForest:
    pass


def yaml_treeify(name:str, yaml:str, *args, **kwargs) -> YamlForest:
    pass


def jsontreeify(name:str, json:str, *args, **kwargs) -> JSONForest:
    pass


# ==================================================================================================
#                                           TEXTIFY
# ==================================================================================================


def header_node_to_text(rootNode: HeaderNode) -> str:
    
    # BASE CASE: We just have text node.
    if isinstance(rootNode, TextNode):
        return rootNode.text + "\n\n"
    
    # For the header node
    headerLevel = rootNode.header_level
    returnString = str(rootNode) + "\n"
    for _, child in enumerate(rootNode.children):
        returnString += header_node_to_text(child)
    
    return returnString    
     

def convert_to_metadata(metadata_dict:dict) -> str:
    yaml = "---\n"
    for key, value in metadata_dict.items():
        
        if isinstance(value, list):
            yaml += f"{key}: \n"
            for elem in value:
                yaml += f"  - {elem}\n"
            continue
        
        yaml += f"{key}: {value}\n"
    yaml += "---\n"
    return yaml


def mdtextify(forest: MarkdownForest) -> str:
    """
    Converts a MarkdownForest to a markdown file.
    """
    
    returnString = ""
    
    # Get the root node
    root = forest.root
    
    # Get the metadata
    metadata = forest.get_metadata()
    returnString = convert_to_metadata(metadata) + "\n"
    
    for node in root.children:
        returnString += header_node_to_text(node) + "\n"
    
    return returnString


def rsttextify(forest: RestructuredForest) -> str:
    
    pass


def asciidoctextify(forest: AsciidocForest) -> str:
    pass


def yamltextify(forest: YamlForest) -> str:
    pass


def jsontextify(forest: JSONForest) -> str:
    pass




from .node import HeaderNode


class Forest:
    
    def __init__(self, root:HeaderNode, documentName:str="[document]", metadata:dict=None):
        self.documentName = documentName
        self.metadata = metadata
        self.root = root
        
    # TODO: Implement __str__ and __len__ methods
    # FIGUREOUT: What other methods should this base class have ?


class MarkdownForest:
    
    # TODO: Make this a subclass of Forest
    def __init__(self, root:HeaderNode, documentName:str="[document]", metadata:dict=None):
        self.documentName = documentName
        self.metadata = metadata
        self.root = root

        # TODO: Implement a method to count the number of nodes in the tree
        
        self.backlinks = []
        self.tags = []
        
    def __str__(self):
        return str(self.root)
    
    def __len__(self):
        return self.treeCount

    def add_root(self, root:HeaderNode):
        self.root = root

    def get_metadata(self):
        return self.metadata
    
    def add_backlink(self, backlink:str):
        self.backlinks.append(backlink)
    
    def get_backlinks(self):
        return self.backlinks
    
    def add_tag(self, tag:str):
        self.tags.append(tag)
        
    def get_tags(self):
        return self.tags
    

# FIGUREOUT: Do we really need a forest for each of the supported formats ?
class RestructuredForest:
    pass


class AsciidocForest:
    pass


class YamlForest:
    pass


class JSONForest:
    pass



from .utils import *

class Parser:
    
    def __init__(self, document_name:str, text:str):
        self.document_name = document_name
        self.text = text
        
    def parse(self) -> Node:
        pass

   
class MarkdownParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        HEADER_PATTERN = r"^(#+\s+)(.*)"
        
        tree = build_tree(self.text, document_name=self.document_name, header_pattern=HEADER_PATTERN)
        return tree
        

class RestructuredParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        HEADER_PATTERN = r"^(=+|-+|`+|'+|\^+|\*+|\.+|~+|\++)\s+(.+)\s+\1$"
        
        tree = build_tree(self.text, document_name=self.document_name, header_pattern=HEADER_PATTERN)
        return tree
        

class TextParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass


class YAMLParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass


class JSONParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass
    

class XMLParser(Parser):
    
    def __init__(self, document_name:str, text:str):
        super().__init__(document_name, text)
        
    def parse(self) -> Node:
        pass
    



from setuptools import setup
from setuptools.command.test import test as TestCommand
from pathlib import Path

this_directory = Path(__file__).parent
LONG_DESCRIPTION = (this_directory / "README.rst").read_text()

VERSION = '1.3.5'
DESCRIPTION = 'A package to convert between markup language documents and a forest data structure for efficient processing.'

setup(
    name = "pydelorean",
    version = VERSION,
    author = "Keane Moraes",
    author_email = 'lordvader3002@gmail.com',
    description = DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    license = "Apache 2.0",
    url = "http://github.com/kj3moraes/delorean",
    packages = ['pydelorean', 'pydelorean.tree', 'pydelorean.parser', 'pydelorean.tools'],
    tests_require = ['unittest'],
    install_requires = ['markdown', 'bigtree', 'beautifulsoup4'],
    download_url = 'https://github.com/kj3moraes/delorean/archive/%s.zip' % VERSION,
    classifiers = [
        "Topic :: Utilities",
    ],
)



import re
from pydelorean.tree import TextNode, HeaderNode
from bigtree import Node, print_tree

def build_tree(inputText:str, *args, **kwargs) -> Node:

    assert isinstance(inputText, str), "inputText must be a string"
    if inputText == "":
        return None
    
    root = HeaderNode(header=kwargs["document_name"], headerNumber=0)

    # Default to the GFM header pattern
    HEADER_PATTERN = kwargs['header_pattern'] if 'header_pattern' in kwargs else r"^(#+\s+)(.*)"
    
    inputText = inputText.strip()
    matches = re.finditer(HEADER_PATTERN, inputText, re.MULTILINE)
        
    # Get all the indices of the headers
    indices = [(match.start(), match.end()) for match in matches]
    
    # BASE CASE
    # Check if it is only text (since there are no header indices)
    if len(indices) == 0:
        text_node = TextNode(name=inputText, text=inputText)
        root >> text_node
        root ** text_node
        return root
    
    # Is there text before the first header?
    if indices[0][0] != 0:
        textBefore = inputText[0:indices[0][0]]
        node = TextNode(name=textBefore, text=textBefore)
        root >> node
        root ** node

    # Use a stack to keep track of the header levels.
    stack = []

    # Iterate over all the headers and accordingly build the tree.
    for currHeader in indices:

        currHeaderText = inputText[currHeader[0]:currHeader[1]]
        currHeaderLevel = len(currHeaderText) - len(currHeaderText.replace("#", ""))
        
        # If this is the first header then create the header node automatically.
        if len(stack) == 0:
            
            # Create the header node with the current header.
            node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
            root >> node
            root ** node
                        
            # Add the node to the stack
            stack.append((currHeaderLevel, currHeader, node))
            continue
        else:
            lastHeader = stack[-1]
        
        # CASE 1: If the current header level is greater than the starting header level then keep going.
        if currHeaderLevel > lastHeader[0]:
            
            # Extract the text between the current header and the previous header
            textBetween = inputText[lastHeader[1][1]:currHeader[0]].strip()
            if textBetween != "":
                node = TextNode(name=textBetween, text=textBetween)
                lastHeader[-1] >> node
                lastHeader[-1] ** node
                
            # Create the header node and add it to the stack
            node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
            lastHeader[-1] >> node
            lastHeader[-1] ** node
            
            stack.append((currHeaderLevel, currHeader, node))
            
        # CASE 2: If the current header level is less than or equal to the starting header level then we need to create a new tree.
        elif currHeaderLevel <= lastHeader[0]:
            
            # Get the text between 
            textBetween = inputText[lastHeader[1][1]:currHeader[0]].strip()
            if textBetween != "":
                node = TextNode(name=textBetween, text=textBetween)
                lastHeader[-1] >> node
                lastHeader[-1] ** node
                                            
            # Pop off until you get to the node with a lower level than the current header level
            while len(stack) > 0:
                if stack[-1][0] >= currHeaderLevel:
                    stack.pop()
                else:
                    break
            
            if len(stack) == 0:
                
                node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
                root >> node
                root ** node
            else:
                node = HeaderNode(header=currHeaderText, headerNumber=currHeaderLevel)
                stack[-1][-1] >> node
                stack[-1][-1] ** node
                
            stack.append((currHeaderLevel, currHeader, node))


    # Check if there is any text after the last header
    if indices[-1][1] != len(inputText):
        textEnd = inputText[indices[-1][1]:].strip()
        node = TextNode(name=textEnd, text=textEnd)
        stack[-1][-1] >> node
        stack[-1][-1] ** node
        
    return root

